/******
 * <<<<<<<<<<<<<<<<<<<<<<<<<=====>>>>>>>>>>>>>>>>>>>>>>>>>
 *    @file:           Val2DDisplay.m  [ $RCSfile: Val2DDisplay.m,v $ ]
 *    @authors:        Ken Cline <Kenneth.C.Cline@cpmx.saic.com>
 *    @date:           Wed Mar  4 16:58:39 EST 1998
 *    @project:        any
 *    @library:        spatial
 *    @package:        NA
 *    @description:    Layer of objects that complement/improve
 *                        those provided in the Swarm "space" lib
 *    @version:        $Revision: 1.4 $,
 *                     $Date: 1998/03/14 23:34:07 $,
 *                     $Author: kcline $
 *    @notes:          NA
 * <<<<<<<<<<<<<<<<<<<<<<<<<=====>>>>>>>>>>>>>>>>>>>>>>>>>
 *****/



#import <spatial/Val2DDisplay.h>
#import <objc/objc-api.h>
#import <string.h>

#ifdef METHODS
   #define displayPoint( x, y, c )   \
      [ displayWidget drawPointX: x Y: y Color: (unsigned char)c ]
#else
      // cache method lookup.
   #define displayPoint( x, y, c )   \
      (void) *drawPointImp( displayWidget, M(drawPointX:Y:Color:), x, y, c )
#endif

#define colorCacheAt( offsets, x, y ) ( (colorCache) + (offsets)[(y)] + (x) )

#define writePt( x, y, z )  \
   [ self writePointX: x Y: y Z: z Dest: displayDestination ]



@implementation Val2DDisplay

+createBegin: (id) aZone {
   Val2DDisplay * newDisplay = [ super createBegin: aZone ];
   id <ProbeMap>  probeMap   = NULL;

   [ newDisplay setDisplayState: Gradient ];

   newDisplay->colorCache = (color_t *)NULL;
   [ newDisplay setCache: OFF ];

   if ( ! [probeLibrary isProbeMapDefinedFor: [self class]] ) {
      probeMap = [ CustomProbeMap create: aZone forClass: [self class]
                                         withIdentifiers: "displayMode",
                                                          "contourModFactor",
                                                          "contourBkgdColor",
                                                          "contourLineColor",
                                                          ":",
                                                          "setDisplayMode:", NULL ];
      [ probeLibrary setProbeMap: probeMap For: [self class] ];
   }

   return newDisplay;
}

-createEnd {
   self = [ super createEnd ];

   if ( contourModFactor == 0 ) contourModFactor = 1;

   return self;
}

-setDiscrete2dToDisplay: (Discrete2d *) grid {
   discrete2d = grid;

   lattice = [ discrete2d getLattice ];
   offsets = [ discrete2d getOffsets ];
   xsize   = [ discrete2d getSizeX   ];
   ysize   = [ discrete2d getSizeY   ];

   return self;
}

-setCache: (BOOL) state {
   useCache = state;

   if ( useCache )
      displayDestination = ARRAY_CACHE;
   else
      displayDestination = RASTER_WIDGET;

   if ( useCache )
      freprintf( stderr, "ValueDisplay: Cache use is `ON'.\n" );

   return self;
}

-setDisplayState: (id <Symbol>) state {
   if ( ! state ) {
      raiseEvent( WarningMessage, "\t> `state' argument is NULL.\n"      \
                                  "\t> Using default display method.\n" );
      displaySelector  = NULL;
      displayMode      = NULL;
      return self;
   }

//   return [ self setDisplayMode: [state getName] ];
   [ self setDisplayMode: [state getName] ];
   return self;
}

-(void) setDisplayMode: (char *) mode {
   id <String> displayMethodName = NULL;
 
   if ( ! mode ) {
      raiseEvent( WarningMessage, "\t> `mode' argument is NULL.\n"      \
                                  "\t> Using default display method.\n" );
      displaySelector  = NULL;
      displayMode      = NULL;
      //return self;
      return;
   }

   displayMethodName = [ String create: [self getZone] setC: "display" ];
   [ displayMethodName catC: (char *)mode ];
   displaySelector = sel_get_uid( [displayMethodName getC] );
 
   if ( [self respondsTo: displaySelector] )
      displayMode = mode;
   else {
      raiseEvent( WarningMessage, "\t> ValueDisplay does NOT"            \
                                     " \"respond to\" method `%s'.\n"    \
                                  "\t> Using default display method.\n", \
                                       [ displayMethodName getC ]        );
      displaySelector  = NULL;
      displayMode      = NULL;
   }

   [ displayMethodName drop ];
   //return self;
   return;
}

-(const char *) getDisplayMode { return displayMode; }


-setContourDisplayMod: (int) m BkgdColor: (Color) bkgdColor
                               LineColor: (Color) lineColor {

   if ( bkgdColor < 0 || bkgdColor > 255 )
      raiseEvent( OutOfRange, NULL, "contourBkgdColor", "[0,255]" );

   if ( lineColor < 0 || lineColor > 255 )
      raiseEvent( OutOfRange, NULL, "contourLineColor", "[0,255]" );

   contourModFactor = m;
   contourBkgdColor = bkgdColor;
   contourLineColor = lineColor;

   return self;
}


-display {
   size_t cacheSize = 0;
   int    x = 0, y = 0;

   if ( ! lattice || ! offsets )
      raiseEvent( WarningMessage, "\t> Value display improperly initialized.\n" );

   if ( ! displaySelector ) return [ super display ];

   if ( useCache ) {
      if ( ! colorCache ) {
         cacheSize = xsize * ysize * sizeof(color_t);
         colorCache = (color_t *)[ [self getZone] allocBlock: cacheSize ];
         memset( colorCache, 0, cacheSize );
         [ self perform: displaySelector ];
      }
      else {
         for ( y = 0 ; y < ysize ; y++ ) {
         for ( x = 0 ; x < xsize ; x++ ) {
            displayPoint( x, y, *colorCacheAt( offsets, x, y) );
         }
         }
      }
      return self;
   }

   return [ self perform: displaySelector ];
}

-displayGradient {
   int  x = 0, y = 0;
   long color = 0;

   for ( y = 0 ; y < ysize ; y++ ) {
   for ( x = 0 ; x < xsize ; x++ ) {

      color = (long)*( discrete2dSiteAt( lattice, offsets, x, y ) );
      color = color / modFactor + colorConstant;
      if ( color < 0 || color > 255 ) {
         raiseEvent( WarningMessage, "\t> Colour %d not in [0,255].\n", color );
      }

      //displayPoint( x, y, color );
      writePt( x, y, color );
   }
   }

   return self;
}

-displayContour {
   int           x = 0,     y = 0,     z = 0;
   gridval_t   z00 = 0,   z10 = 0,   z01 = 0, z11 = 0;
   int         m00 = 0,   m10 = 0,   m01 = 0, m11 = 0;
   int       d1100 = 0, d1110 = 0, d1101 = 0;

   for ( y = 0 ; y < ysize - 1 ; y++ ) {

      z10 = (long)*(discrete2dSiteAt( lattice, offsets, x, y + 0 ));
      z11 = (long)*(discrete2dSiteAt( lattice, offsets, x, y + 1 ));
      m10 = z10 / contourModFactor;
      m11 = z11 / contourModFactor;

   for ( x = 0 ; x < xsize - 1 ; x++ ) {

      z00 = z10; z10 = (long)*(discrete2dSiteAt( lattice, offsets, x + 1, y + 0 ));
      z01 = z11; z11 = (long)*(discrete2dSiteAt( lattice, offsets, x + 1, y + 1 ));
      m00 = m10; m10 = z10 / contourModFactor;
      m01 = m11; m11 = z11 / contourModFactor;

      d1100 = abs( m11 - m00 );
      d1110 = abs( m11 - m10 );
      d1101 = abs( m11 - m01 );

      if ( d1100 || d1110 || d1101 ) {
         writePt( x + 1, y + 1, contourLineColor );
         if ( d1100 > 1 ) writePt( x + 0, y + 0, contourLineColor );
         if ( d1110 > 1 ) writePt( x + 1, y + 0, contourLineColor );
         if ( d1101 > 1 ) writePt( x + 0, y + 1, contourLineColor );
      }
      else {
         if ( contourBkgdColor == -1 ) {
            z = z11 / modFactor + colorConstant;
            if ( z < 0 || z > 255)
               raiseEvent( WarningMessage, "\t> Colour %d not in [0,255].\n", z );
            writePt( x + 1, y + 1, z );
         }
         else
            writePt( x + 1, y + 1, contourBkgdColor );
      }
   }
   }
   return self;
}

-displayContourOverlay {
   int tempColor = contourBkgdColor;

   contourBkgdColor = -1;             // -1 => "transparent"
   [ self displayContour ];
   contourBkgdColor = tempColor;

   return self;
}


-(void) writePointX: (int) x Y: (int) y Z: (int) z Dest: (int) dest {
   switch ( dest ) {
      case ARRAY_CACHE :
         *colorCacheAt( offsets, x, y ) = (color_t)z;
         displayPoint( x, y, z );
         break;
      case RASTER_WIDGET :
         displayPoint( x, y, z );
         break;
      default :
         displayPoint( x, y, z );
         break;
   }
}


@end



